/*
 *  Copyright(c)2002 Forward Computing and Control Pty. Ltd.
 *  ACN 003 669 994, NSW, Australia      All rights Reserved
 *  
 *   Written by Dr. M.P. Ford
 *
 *  ThreadReturn package Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *     copyright notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *     include, in the documentation and/or other materials provided with the 
 *     distribution, a statement that the software has been modified and
 *     identify the person who was responsable for the modifications.  
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
package au.com.forward.threads;
import java.util.EventObject;
import java.nio.channels.ClosedByInterruptException;
import java.io.Serializable;


/**
 * This class encapsulates a <code>Thread</code> termination event.&nbsp;
 * See {@link ThreadReturn} for more details.
 * <p>
 * This class contains the source thread, the type of termination and the return <code>Object</code>
 * or <code>Throwable</code>.&nbsp It is used as the argument to the {@link ThreadListener} methods.
 * <p>
 *
 * Mods: 8th March 2003 - added implements Serializable added processing for Object arg.
 *
 * @author     matthew ford 
 * @created    7 April 2002 
 */
public class ThreadEvent extends EventObject implements Serializable {

	/**
	 *  This type indicates the thread died normally.&nbsp;
	 * <p>
	 *  </code>{@link #getObject} returns the object saved by this thread, if any.<br>
	 * {@link #getThrowable} will return <code>null</code>. 
	 * @see #getType
	 */
	public final static int RETURN = 0;

	/**
	 *  This type indicates the thread saved either an <code>InterruptedException</code> or
	 * a <code>ClosedByInterruptException</code>.&nbsp;
	 * <p>
	 *  {@link #getThrowable} returns either <code>InterruptedException</code> or
	 *   <code>ClosedByInterruptException</code><br>
	 *  {@link #getObject} returns the exception as an <code>Object</code>.
	 * @see #getType
	 */
	public final static int INTERRUPTED = 1;

	/**
	 *  This type indicates there was some error in the thread execution or monitoring.&nbsp;
	 * <p>
	 *  {@link #getThrowable} returns the associated <code>Throwable</code> as a <code>Throwable</code>.&nbsp;<br>
	 *  {@link #getObject} returns the associated <code>Throwable</code> as an <code>Object</code>.
	 * @see #getType
	 */
	public final static int ERROR = 2;

	/**
	 *  Type of this event 
	 */
	private int type = RETURN;

	/**
	 *  Object associated with this event 
	 */
	private Object object = null;


	/**
	 *  The thread died normally and saved this object for return.&nbsp; 
	 * {@link #getType()} will return {@link #RETURN} 
	 *
	 * @param  source  the <code>Thread</code> being monitored. 
	 * @param  obj     the <code>Object</code> saved by the thread or <code>null</code> if none saved. 
	 */
	public ThreadEvent(Thread source, Object obj) {
		super(source);
		if (obj instanceof Throwable) {
			if ((obj instanceof InterruptedException) ||
					(obj instanceof ClosedByInterruptException) ) {
					type = INTERRUPTED;
			} else {
				type = ERROR;
			}
		}	
		object = obj;
	}


	/**
	 *  The thread saved some <code>Throwable</code> or some error occurred
	 *  during the  monitoring.&nbsp; 
	 * {@link #getType()} will return {@link #INTERRUPTED} if the <code>Throwable</code>
	 * is either <code>InterruptedException</code> or
	 * <code>ClosedByInterruptException</code> otherwise it will return {@link #ERROR} 
	 *
	 * @param  source  the <code>Thread</code> being monitored. 
	 * @param  e       the <code>Throwable</code> describing the error.
	 */
	public ThreadEvent(Thread source, Throwable e) {
		super(source);
		if ((e instanceof InterruptedException) ||
			  (e instanceof ClosedByInterruptException) ) {
			  type = INTERRUPTED;
	  } else {
			type = ERROR;
	  }	
		object = e;
	}



	/**
	 *  Gets the type of this {@link ThreadEvent}
	 *
	 * @return    the type, {@link #RETURN}, {@link #INTERRUPTED} or {@link #ERROR}
	 */
	public int getType() {
		return type;
	}

	/**
	 *  Gets the source <code>Thread</code> of this event. 
	 *
	 * @return    the <code>Thread</code> being monitored.
	 */
	public Thread getThread() {
		return (Thread) getSource();
	}


	/**
	 *  Gets the <code>Throwable</code> object saved by the thread.&nbsp;
	 *  Returns <code>null</code> if the thread did not save a <code>Throwable</code> object. 
	 *
	 * @return    the <code>Throwable</code> object or <code>null</code> if none saved.
	 */
	public Throwable getThrowable() {
		if (object instanceof Throwable) {
			return (Throwable) object;
		}
		// else
		return null;
	}


	/**
	 *  Gets the <code>Object</code> saved by the thread.&nbsp;
	 *  If the thread saved a <code>Throwable</code> object that object will be returned. 
	 *
	 * @return    The <code>Object</code> saved by the thread or <code>null</code> if none saved. 
	 */
	public Object getObject() {
		return object;
	}


	/**
	 *  Returns a string containing a description of this {@link ThreadEvent}.
	 *
	 * @return    description of this {@link ThreadEvent}.
	 */
	public String toString() {
		Thread thread = getThread();
		
		String rtn = "ThreadEvent_source:" + ((thread==null)?"null":thread.getName())+"\n"+
		          " ThreadEvent_type:"; 
		switch(type) {
			case RETURN: rtn+= "RETURN"; break;
			case INTERRUPTED: rtn += "INTERRUPTED"; break;
			default: rtn += "ERROR"; break;
		}	
		rtn += "\n";
	
		if (object == null) {		
			rtn += " Stored_return:null";
		} else {	
			if (object instanceof Throwable) {
				rtn += " Stored_exception:" + StackTrace.toString((Throwable) object);
			} else {
				rtn += " Stored_return:" + object.toString();
			}
		}	
		return rtn;
	}
}