/*
 *  Copyright(c)2002 Forward Computing and Control Pty. Ltd.
 *  ACN 003 669 994, NSW, Australia      All rights Reserved
 *  
 *   Written by Dr. M.P. Ford
 *
 *  ThreadReturn package Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *     copyright notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *     include, in the documentation and/or other materials provided with the 
 *     distribution, a statement that the software has been modified and
 *     identify the person who was responsable for the modifications.  
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
package au.com.forward.threads;
import java.util.WeakHashMap;
import java.lang.ref.WeakReference;
import java.util.Map;
import java.util.ArrayList;
import java.util.Collections;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.nio.channels.ClosedByInterruptException;

/*
 * Mods V1.0.3 2005-05-11
 * Changed test in addListener() to == instead of .contains() which uses .equals() and so may
 * not add listeners that are actually different objects.
 */
/**
 *  This is the main class of this package and contains the static methods for
 *  handling threads.
 * <p>
 * This class contains a number 
 * of <code>static</code> methods that allow you to
 * <ul>
 * <li>return a result from a thread, {@link au.com.forward.threads.ThreadReturn#save ThreadReturn.save(Object)} and {@link au.com.forward.threads.ThreadReturn#join ThreadReturn.join()}</li>
 * <li>catch exceptions thrown by a thread and re-throw them in the 'calling' thread, 
 * {@link au.com.forward.threads.ThreadReturn#save ThreadReturn.save(Object)} and {@link 
 * au.com.forward.threads.ThreadReturn#join ThreadReturn.join()}</li>
 * <li>stop a thread by calling its <code>interrupt()</code> method, {@link au.com.forward.threads.ThreadReturn#ifInterruptedStop ThreadReturn.ifInterruptedStop()}</li>
 * <li>add {@link au.com.forward.threads.ThreadListener}s threads to be called when the thread has died,
 *  {@link au.com.forward.threads.ThreadReturn#addListener ThreadReturn.addListener(Thread, ThreadListener)}</li>
 * </ul>
 *
 * <h2>Copyright</h2>
 * You may not re-distribute the source code in any form.&nbsp;
 * <p>
 * Provided you add the following acknowledgement
<pre>
    This program incorporates code provided by
    Forward Computing and Control Pty. Ltd.,
    www.forward.com.au
</pre> 
 *  you may re-distribute the compiled java .class file in both private and commercial products.
 *
 * <h2>Guidelines for writing the run() method</h2>
 * The effectiveness of this class depends on the use of an outer <code>try/catch</code> block in the
 * thread's <code>run()</code> method and the addition of {@link ThreadReturn#save ThreadReturn.save(Object)}
 * and {@link ThreadReturn#ifInterruptedStop ThreadReturn.ifInterruptedStop()} statements.&nbsp;
 * <p>
 * Here is an example of a <code>Thread</code> class <code>run()</code> method that follows this outline.
 <pre><code>
public class ExampleThread {
  
  public void run() {
    // put a try/catch at the outermost level
    try {
      // your thread code goes here
      // do some work ...
      // ...
      
      // stop this thread if it has been interrupted
      ThreadReturn.ifInterruptedStop();
      
      // do some more work ...
      // ...
      
      // stop this thread if it has been interrupted
      ThreadReturn.ifInterruptedStop();
      
      //  .....
      
      ThreadReturn.save(result);  // optional if you want to return a result
    } catch (Throwable e) {
      // catch all exceptions, errors etc. thrown by this thread here
      ThreadReturn.save(e);
    }
  }
} 
</code></pre>
* As well as, this you must insure that all fatal errors that occur in methods called by the <code>run()</code>
* method throw an exception back up to the <code>run()</code> method to be caught by the<br>
* &nbsp;&nbsp;&nbsp;&nbsp;<code>} catch (Throwable e) {</code><br>
* clause.
* <p>
*
* Then in the parent thread, you can use the following code to wait until the sub-thread dies and collect
* the result.

 <pre><code>
Object threadResult = null;
Thread t1 = new ExampleThread();
t1.start();
try {
  threadResult = ThreadReturn.join(t1);
} catch (InterruptedException ex) {
  // some error running t1, handle it here
  if (ex instanceof ThreadException) {
    // t1 throw a Throwable which is the cause of this ThreadException
  } else {
    // ThreadReturn.join(t1) was interrupted
  }
}
</code></pre>
* Like the <code>Thread</code> class <code>join()</code> method that <code>ThreadReturn.join(t1)</code> is replacing,
* <code>ThreadReturn.join(t1)</code> can throw an <code>InterruptedException</code>.&nbsp; If the thread <code>t1</code>'s <code>run()</code>
* terminated because a <code>Throwable</code> was caught in its <code>catch (Throwable e)</code> clause,
* then <code>ThreadReturn.join(t1)</code> will throw a {@link ThreadException} with that 
* <code>Throwable</code> stored as it cause.&nbsp; <code>ex.getCause()</code> will retrieve it or you can 
* use <code>StackTrace.toString(ex)</code> to get the whole stacktrace as a <code>String</code>.
*<p>
* The result returned by <br>&nbsp;&nbsp;&nbsp;&nbsp;<code>threadResult = ThreadReturn.join(t1);</code><br> is the <code>Object</code>
* saved by the <br>&nbsp;&nbsp;&nbsp;&nbsp;<code>ThreadReturn.save(result);</code><br> statement at the end of the <code>run()</code>
* method (just above the <br>&nbsp;&nbsp;&nbsp;&nbsp;<code>} catch (Throwable e) {</code><br> line.
* <p>
* If you did not put this statement in your <code>run()</code> method
* <code>ThreadReturn.join(t1);</code> will return <code>null</code>
* <p>
* To wait at most <i>x</i> milliseconds for the sub-thread to time out you can use one of the two versions of 
* <code>ThreadReturn.join</code>
* that specify a timeout, {@link #join(Thread, long) ThreadReturn.join(Thread, long)} or 
* {@link #join(Thread, long, int) ThreadReturn.join(Thread, long, int)} e.g. to wait at
* most <code>500</code> milliseconds
 <pre><code>
Object threadResult = null;
Thread t1 = new ExampleThread();
t1.start();
try {
  threadResult = ThreadReturn.join(t1, 500);
} catch (InterruptedException ex) {
  // some error running t1, handle it here
  if (ex instanceof ThreadException) {
    // t1 throw a Throwable which is the cause of this ThreadException
  } else if (ex instanceof TimedOutdException) {
    // join timedout before t1 died
  } else {
    // ThreadReturn.join(t1) was interrupted
  }
}
</code></pre>
* In this case <code>ThreadReturn.join(t1,500)</code> can also throw a {@link TimedOutException} to indicate
* the thread did not finish in the specified time.

* <h2>To terminate a thread</h2>
* To terminate a thread early call <code>ThreadReturn.stop(thread);</code>.&nbsp;
* If the thread in is not blocked then its interrupted status will be set and the next time it
* executes a {@link #ifInterruptedStop ThreadReturn.ifInterruptedStop()} statement an <code>InterruptedException</code>
* will be thrown causing the <code>run()</code> method to terminate.
* If the thread is waiting in a <code>wait()</code>, <code>join()</code> or <code>sleep()</code>
* then an <code>InterruptedException</code> will be thrown.
* If the thread is blocked in an I/O operation on a <code>java.nio.channels.interruptibleChannel</code>
* then a <code>ClosedByInterruptException</code> will be thrown.
* Any of these cases will cause the <code>run()</code> method to terminate.
* <p>
* Using <code>ThreadReturn.stop(thread);</code> instead of <code>thread.interrupt();</code> has the following
* advantages.<br>
* i)<code>thread.interrupt();</code> will not set the interrupt flag for threads that have not started yet.<br>
* ii)<code>ThreadReturn.stop(thread);</code> will continue to interrupt the thread until it dies. This
* overcomes problems where a method called by the thread's <code>run()</code> method catches and ignores the first <code>interrupt()</code> call.
* <p> 
* All the versions of <code>ThreadReturn.join()</code> call the thread's <code>interrupt()</code>
* method before they return, even if an exception is being thrown.&nbsp;  This means that if you have written
* your thread's <code>run()</code> method according to the guidelines above your thread will be terminated
* if <code>ThreadReturn.join()</code> throws any exception.
* 
* <h2>ThreadListeners</h2>
* You can add one or more {@link ThreadListener}s to a thread using the {@link ThreadReturn#addListener 
* ThreadReturn.addListener(Thread, ThreadListener)} method.  Classes implementing the {@link ThreadListener}
* interface have three methods {@link ThreadListener#threadResult ThreadListener.threadResult(ThreadEvent)}, 
* {@link ThreadListener#threadInterrupted ThreadListener.threadInterrupted(ThreadEvent)} 
*  and {@link ThreadListener#threadError ThreadListener.threadError(ThreadEvent)}.
*<p>
* If the thread terminates normally, {@link ThreadListener#threadResult threadResult(ThreadEvent)}
* is called with a {@link ThreadEvent} containing the source thread, a type of {@link ThreadEvent#RETURN} and
* the result <code>Object</code> saved by the thread, if any.&nbsp;
* If the {@link ThreadListener#threadResult threadResult(ThreadEvent)} method
* throws any <code>Throwable</code> object, it is passed onto that listener's 
*{@link ThreadListener#threadError threadError(ThreadEvent)} method.&nbsp; Errors in one listener's
* {@link ThreadListener#threadError ThreadListener.threadError(ThreadEvent)} method do not effect the
* processing of other listeners registered for this thread.<br>
* <b>NOTE:</b>{@link ThreadListener#threadResult threadResult(ThreadEvent)} should not modify the
* result <code>Object</code> contained in the {@link ThreadEvent} because the same result <code>Object</code>
* will be passed on to the other registered listeners for this thread. 
* 
*<p> 
* If the thread terminates and has saved an <code>InterruptedException</code> or a 
* <code>ClosedByInterruptException</code>, then 
* {@link ThreadListener#threadInterrupted threadInterrupted(ThreadEvent)}
* is called with a {@link ThreadEvent} containing the source thread, a type of {@link ThreadEvent#INTERRUPTED} and
* the <code>Exception</code> saved by the thread.&nbsp;
* <p>
* If the {@link ThreadListener#threadInterrupted threadInterrupted(ThreadEvent)} method
* throws any <code>Throwable</code> object, it is passed onto that listener's 
*{@link ThreadListener#threadError threadError(ThreadEvent)} method.&nbsp; Errors in one listener's
* {@link ThreadListener#threadError ThreadListener.threadError(ThreadEvent)} method do not effect the
* processing of other listeners registered for this thread.<br>
* <b>NOTE:</b>{@link ThreadListener#threadInterrupted threadInterrupted(ThreadEvent)} should not modify the
* result <code>Object</code> contained in the {@link ThreadEvent} because the same <code>Exception</code>
* will be passed on to the other registered listeners for this thread. 
* 
* <p> 
* If the thread terminates and has saved some other type of <code>Throwable</code> or if the system monitor thread
* is interrupted or there is some other error, then {@link ThreadListener#threadError threadError(ThreadEvent)}
* is called with a {@link ThreadEvent} containing the source thread, a type of {@link ThreadEvent#ERROR} and
* the <code>Throwable</code> object that was thrown.&nbsp;
* If the {@link ThreadListener#threadError threadError(ThreadEvent)} method
* throws any <code>Throwable</code> objects, they are caught and ignored.&nbsp;
* It is your responsibility to catch and handle <em><b>ALL</b></em> 
*  <code>Throwable</code> objects thrown by this method. 
* <p>
* {@link ThreadListener}s can be added (or removed) at any time.&nbsp;
* <p>
* If a {@link ThreadListener}
* is added before the thread is started, the thread is added to the list which is checked every 
* {@link #getCheckDelay()} milliseconds.&nbsp; When the thread has started (or has died), a monitor
* thread is started and joins the thread waiting for it to finish.  When the thread has died,
* all the registered {@link ThreadListener}s are called and removed.
* <p>
* Because a thread can only be run once, each listener is removed when it is called.
* <p>
* If a {@link ThreadListener} is added after the thread has started (or after it has died), a monitor
* thread is started and joins the thread waiting for it to finish.  When the thread has died,
* all the registered {@link ThreadListener}s are called and removed.
* <p>
* <b>NOTE:</b><br>
* i) All {@link ThreadListener} methods are called on a monitor thread that is neither
* the thread being listened to nor the thread that added the listener. <br>
* ii) If you add a {@link ThreadListener} after the thread has died, the monitor thread will
* call the {@link ThreadListener} as soon as it starts and passes the {@link ThreadListener},
* either the <code>Object</code>
* the thread returned, or the <code>Throwable</code> that was thrown.
*
* <h2>Thread resources</h2>
* If you don't use the {@link ThreadReturn#addListener ThreadReturn.addListener(Thread,ThreadListener)}
* method, no additional threads are started by this package.&nbsp;  However when you add a {@link ThreadListener}, a
* monitor thread is started for each thread that has listeners.&nbsp;  This monitor thread calls 
* {@link ThreadReturn#join} to detect when the thread has died and then calls the {@link ThreadListener}s
* registered for that thread.&nbsp;  The monitor thread then dies.  
* <p>
* If you add a {@link ThreadListener} to a thread that has not started then a single daemon check thread is 
* started.&nbsp;  Once started this check thread continues to run until the application terminates.&nbsp;
* The check thread goes to <code>sleep</code> for the number of milliseconds set by {@link #setCheckDelay} 
* (default 100mS) and then checks
* if any of the threads with listeners have started since the last check.&nbsp;  For any that have started, or have
* started and died, since the last check, the check thread starts a monitor thread to handle that thread's 
* listeners.&nbsp;  It then loops to <code>sleep</code> again.
*
* <h2>Debug flag</h2>
* If you set the {@link ThreadReturn#debug ThreadReturn.debug} to <code>true</code>, debugging messages
* will be sent to <code>System.out</code>.
*
* <h2>saveImmedeately flag</h2>
*  Set {@link ThreadReturn#saveImmedeately ThreadReturn.saveImmedeately} to <code>true</code> 
*  to force listeners to be called as soon as Thread.save()
*  is called for the first time without waiting for the thread to die and the check thread to notice it.
*  The listeners will be called in the running thread and then removed so subsequent calls
*  to Thread.save() from this thread will NOT fire listeners.  However the final result stored
*  will be that from the last Thread.save(), so only have one Thread.save() at the bottom of the
*  try{} block and one other in the catch(){} block, to ensure consistent results.
*
* @author     matthew ford 
* @created    7 April 2002
*  
* Updated 2nd Nov. 2004
* Added saveImmedeately flag.
*/
public class ThreadReturn {

	/**
	 *  Set this flag to <code>true</code> to output debugging messages to <code>System.out</code> 
	 */
	public static boolean debug = false;
	
	/**
	 *  Set this flag to <code>true</code> to force listeners to be called as soon as Thread.save()
   *  is called for the first time without waiting for the thread to die and the check thread to notice it.
   *  The listeners will be called in the running thread and then removed so subsequent calls
   *  to Thread.save() from this thread will NOT fire listeners.  However the final result stored
   *  will be that from the last Thread.save(), so only have one Thread.save() at the bottom of the
   *  try{} block and one other in the catch(){} block, to ensure consistent results.
	 */
	public static boolean saveImmedeately = false;

	/**
	 *  Stores the threads which there are listeners for but which have not yet 
	 *  started.  Need to synchronize on this object before accessing it 
	 */
	static ArrayList threadsNotStarted = new ArrayList();

	/**
	 *  Stores the threads which have been stopped but have not yet returned 
	 *  Need to synchronize on this object before accessing it 
	 */
	static ArrayList threadsStopped = new ArrayList();
	
	/**
	 *  The delay in the loop the checks if threads have started. Default 100mS 
	 */
	private static long checkDelay = 100;

	/**
	 *  Stores the first exception thrown in a thread under its thread object. The 
	 *  methods are synchronized so it is safe to use in this multi-threaded 
	 *  environment. However need to synchronize on thread when doing update of Map.
	 */
	private static Map threadReturnsMap = Collections.synchronizedMap(new WeakHashMap());

	/**
	 *  Stores the listener list for each thread under its thread object. The 
	 *  methods are synchronized so it is safe to use in this multi-threaded 
	 *  environment. However need to synchronize on thread when doing update of Map.
	 */
	private static Map threadListenerMap = Collections.synchronizedMap(new WeakHashMap());

	/**
	 *  Stores the list of listeners that have been called for each thread under its thread object. The 
	 *  methods are synchronized so it is safe to use in this multi-threaded 
	 *  environment. However need to synchronize on thread when doing update of Map.
	 */
	private static Map calledListenerMap = Collections.synchronizedMap(new WeakHashMap());

	/**
	 *  Stores the ThreadJoin thread for each thread under its thread object. The 
	 *  methods are synchronized so it is safe to use in this multi-threaded 
	 *  environment. However need to synchronize on thread when doing update of Map.
	 */
	private static Map threadJoinMap = Collections.synchronizedMap(new WeakHashMap());

	/**
	 *  Description of the Field 
	 */
	private static SimpleDateFormat dateAndTime = new SimpleDateFormat("yy/MM/dd HH:mm:ss.SS");


	/**
	 *  Thread to check threads waiting to start 
	 */
	private static Thread checkThread = null;


 /**
  * start the CheckThread if a ThreadListener is added for a thread that has not been started yet.
  */
	private static synchronized void startCheckThread() {
		// start up thread to check for threads not started yet
		if (checkThread != null) {
			// already started by some other thread
			return;
		}
			
		Thread t = 
			new Thread("au.com.forward.threads.checkThread") {
				/**
				 *  Main processing method for the ThreadReturn object 
				 */
				public void run() {
					Thread thread;
					while (true) {
						try {
							try {
								sleep(checkDelay);
							} catch (InterruptedException ex) {
								// ignore
							}
							synchronized (threadsNotStarted) {
								if (debug && (threadsNotStarted.size() > 0)) {
									System.out.println("ThreadReturn debug:" + dateAndTime.format(new Date()) + " Checking for threads started.");
								}
								for (int i = threadsNotStarted.size()-1; i >= 0; i-- ) {
									// start from size and work down because remove(i) will move up other threads in list
									thread = (Thread) ((WeakReference)threadsNotStarted.get(i)).get(); // get weakrefernce object
									if (thread == null) { 
										// this thread has been garbage collected
										threadsNotStarted.remove(i);
										continue;
									} 	
										
									if (hasDied(thread)) {
										// thread has died just call join on a utility thread
										threadsNotStarted.remove(i);
										if (debug) {
											System.out.println("ThreadReturn debug:" + "                 " + thread.getName() + " started and died since last check.");
										}
										joinThread(thread);
									} else {
										// thread has not died
										if (thread.isAlive()) {
											// thread running just call join
											threadsNotStarted.remove(i);
											if (debug) {
												System.out.println("ThreadReturn debug:" + "                 " + thread.getName() + " started since last check.");
											}
											joinThread(thread);
										} else {
											// need to wait until this thread to start
											if (debug) {
												System.out.println("ThreadReturn debug:" + "                 " + thread.getName() + " not started since last check.");
											}
										}
									}
								}
							}
							// now check threads stopped
							synchronized (threadsStopped) {
								if (debug && (threadsStopped.size() > 0)) {
									System.out.println("ThreadReturn debug:" + dateAndTime.format(new Date()) + " Checking for threads stopped.");
								}
								for (int i = threadsStopped.size()-1; i >= 0; i-- ) {
									// start from size and work down because remove(i) will move up other threads in list
									Thread thread_i = (Thread) ((WeakReference)threadsStopped.get(i)).get(); // get weakrefernce object
									if (thread_i == null) { 
										// this thread has been garbage collected
										threadsStopped.remove(i);
										if (debug) {
											System.out.println("ThreadReturn debug:" + "                 " + "Thread at index:"+i + " has been garbage collected. Removed it from stop list.");
										}
										continue;
									} 
									// else found one still alive interrupt it again
									if (!hasDied(thread_i)) {
										if (debug) {
											System.out.println("ThreadReturn debug:" + "                 " + thread_i.getName() + " was stopped but has not yet died. Interrupt it again.");
										}
										thread_i.interrupt();
									} else {
										// has died remove it
										threadsStopped.remove(i);
										if (debug) {
											System.out.println("ThreadReturn debug:" + "                 " + thread_i.getName() + " has died. Removing from stop list.");
										}
									}	
								}	
							}

						} catch (Throwable t) {
							// ignore
						}
					}
					// loop forever
				}
			};
		t.setDaemon(true);
		checkThread = t;
		if (debug) {
			System.out.println("ThreadReturn debug:" +  dateAndTime.format(new Date()) + " starting thread "+t.getName()+" to check for threads starting.");
		}
		t.start();
	}


	/**
	 *  No objects of this type are ever created 
	 */
	private ThreadReturn() {
	}


	/**
	 *  Sets the loop delay for the check thread that checks on threads waiting to be started.&nbsp; Default 
	 *  100mS. 
	 *
	 * @param  millisecs  the delay, in milliseconds, between checking if threads with listeners have been stared. 
	 */
	public static void setCheckDelay(long millisecs) {
		checkDelay = millisecs;
	}


	/**
	 *  Gets the loop delay for the check thread that checks on threads waiting to be started.&nbsp; Default 
	 *  100mS. 
	 *
	 * @return    the delay, in milliseconds, between checking if threads with listeners have been stared.
	 */
	public static long getCheckDelay() {
		return checkDelay;
	}


	/**
	 *  Checks if this thread has died 
	 *
	 * @return true if thread has died
	 */
	public static boolean hasDied(Thread thread) {
		return (thread.getThreadGroup() == null);
	}
	
	/**
	 *  Adds a listener for a thread.&nbsp; This {@link ThreadListener} is only called once, when the
	 * thread has died, and then discarded, because threads can only be run once.
	 * The same listener cannot be added twice for the same thread.
	 * <p>
	 *  All calls to this listener are made by a separate monitor thread.  I.e.
	 * the monitor thread is neither the <code>thread</code> being listened to, nor the thread calling this method.
	 *  <p>
	 * <ul>
	 *  <li>If the thread has already 
	 *  died, the listener is called immediately by the monitor thread and discarded.&nbsp;
	 * <li> If the thread is 
	 *  running, the monitor thread joins it and waits for it to die
	 *  and then calls the listener and discards it.
	 * <li>
	 * If the thread has not started yet, the thread is added to the list of threads
	 *  to be checked every {@link #getCheckDelay} milliseconds to see when it starts.
	 *  When the thread has started, the monitor thread joins it and waits for it to die
	 *  and then calls the listener and discards it.
	 * </ul>
	 *
	 * @param  thread        the thread to listen for its death.
	 * @param  listener      the {@link ThreadListener} to call when the
	 * thread has died.
	 */
	public static void addListener(Thread thread, ThreadListener listener) {
		// add listener to this thread
		if (hasDied(thread) && thread.isAlive()) {
			// thread has died and start() was called again
			throw new IllegalThreadStateException(thread.getName() + " was re-started after it had died.");
		}

		if (checkThread == null) {
			// this method is synchronized and checks for checkThread != null 
			// so only one check thread will be started.
			startCheckThread();
		}			
		synchronized (thread) {
			// lock this thread's listeners against modification by other threads
			ArrayList listenerList = (ArrayList) threadListenerMap.get(thread);
			if (listenerList == null) {
				listenerList = new ArrayList();
			}
			
			ArrayList calledListenerList = (ArrayList) calledListenerMap.get(thread);
			if (calledListenerList == null) {
				calledListenerList = new ArrayList();
				calledListenerMap.put(thread, calledListenerList);
			}
			
			// test was contains() but this give in-correct results if equals has been overridden.
			for (int i=0; i<listenerList.size(); i++) {
				if (listenerList.get(i) == listener) {
				// already added
					return;
				}
			}	
			for (int i=0; i<calledListenerList.size(); i++) {
				if (calledListenerList.get(i) == listener) {
				// already added
					return;
				}
			}	
			// else not already there add it
			listenerList.add(listener);
			threadListenerMap.put(thread, listenerList);
			if (debug) {
			    System.out.println("ThreadReturn debug: Added Listener "+listener);
			}
			if (hasDied(thread)) {
				// thread has died just call join on a utility thread
				joinThread(thread);
			} else {
				// thread has not died
				if (thread.isAlive()) {
					// thread running just call join
					joinThread(thread);
				} else {
					// need to wait until this thread to start
					// use WeakReference to allow thread to be garbaged collected if it is never started.
					synchronized (threadsNotStarted) {
						threadsNotStarted.add(new WeakReference(thread));
					}
				}
			}
		}
	}


	/**
	 *  Removes a listener from a thread, if the listener has not already been 
	 *  removed. 
	 *
	 * @param  thread    the thread to remove the listener from. 
	 * @param  listener        the {@link ThreadListener} to remove. If the 
	 *  {@link ThreadListener} is no longer registered for this <code>thread</code>, this method
	 * just returns.
	 */
	public static void removeListener(Thread thread, ThreadListener listener) {
		synchronized (thread) {
			// lock this thread's listeners against modification by other threads
			ArrayList listenerList = (ArrayList) threadListenerMap.get(thread);
			if (listenerList == null) {
				return;
			}
			listenerList.remove(listener);
			threadListenerMap.put(thread, listenerList);
			if (debug) {
			    System.out.println("ThreadReturn debug: Removed Listener "+listener);
			}
			
		}
	}


	/**
	 *  Sets the stop flag for this thread in the <code>threadStopMap</code> and calls the
	 *  thread's <code>interrupt()</code> method.
	 *  Then each time the listener thread runs interrupt is called again until the thread dies
	 *
	 */
	public static void stop(Thread thread){
		if (hasDied(thread)) {
			// thread has died
			if (thread.isAlive()) {
				// start() was called again
				throw new IllegalThreadStateException(thread.getName() + " was re-started after it had died.");
			}
			// else has died the first time no need to interrupt
			return;
		} // else  thread has not died
		if (thread.isAlive()) {
			thread.interrupt(); // interrupt it
		}	
		boolean found = false;
		synchronized (threadsStopped) {
			for (int i = threadsStopped.size()-1; i >= 0; i-- ) {
				// start from size and work down because remove(i) will move up other threads in list
				Thread thread_i = (Thread) ((WeakReference)threadsStopped.get(i)).get(); // get weakrefernce object
				if (thread_i == null) { 
					// this thread has been garbage collected
					threadsStopped.remove(i);
					continue;
				} 
				// else
				if (thread.equals(thread_i)) {
					found = true;
					break;
				}
			}
			if ( (!found) && (!hasDied(thread)) ) {
				// not there and still alive
				threadsStopped.add(new WeakReference(thread));
			}	
		}
		if (checkThread == null) {
			// this method is synchronized and checks for checkThread != null 
			// so only one check thread will be started.
			startCheckThread();
		}			
	}

	
	/**
	 *  Calls <code>yield()</code> on the current thread and then throws an <code>InterruptedException</code>
	 * if the thread's interrupted state is set.&nbsp; 
	 *  This method does not clear the thread's interrupt state. 
	 *
	 * @throws InterruptedException	 if current thread's 
	 *      interrupt state is set. 
	 */
	public static void ifInterruptedStop() throws InterruptedException {
		//Thread.yield();
		// let another thread have some time, perhaps to stop this one.
		Thread thread = Thread.currentThread();
		if (thread.isInterrupted()) {
			throw new InterruptedException();
		} // else check is this thread has been stopped
		synchronized (threadsStopped) {
			for (int i = threadsStopped.size()-1; i >= 0; i-- ) {
				// start from size and work down because remove(i) will move up other threads in list
				Thread thread_i = (Thread) ((WeakReference)threadsStopped.get(i)).get(); // get weakrefernce object
				if (thread_i == null) { 
					// this thread has been garbage collected
					threadsStopped.remove(i);
					continue;
				} 
				// else
				if (thread.equals(thread_i)) {
					// yes it was stopped because it is on the list
					thread.interrupt(); // set interrupt flag
					throw new InterruptedException();
				}
			}
		}	
	}


	/**
	 *  Saves this <code>Object</code> against the current thread.&nbsp;
	 * <p>
	 *  If the object saved is not a <code>Throwable</code>,
	 *  then the object will be returned by the {@link #join ThreadReturn.join()} methods and will be
	 *  passed in a {@link ThreadEvent} to the {@link ThreadListener#threadResult ThreadListener.threadResult(ThreadEvent)} method of 
	 *  {@link ThreadListener}s registered for the thread.
	 * <p>
	 *  If the object is an <code>InterruptedException</code> or a 
	 * <code>ClosedByInterrupteException</code>, then the {@link #join ThreadReturn.join()} methods
	 *  will throw a {@link ThreadInterruptedException} and the object saved will be
	 *  passed in a {@link ThreadEvent} to the {@link ThreadListener#threadInterrupted ThreadListener.threadInterrupted(ThreadEvent)} method of 
	 *  {@link ThreadListener}s registered for the thread.
	 * <p>
	 *  If the object saved is any other <code>Throwable</code>, then the {@link #join ThreadReturn.join()} methods
	 *  will throw a {@link ThreadException} and the <code>Throwable</code> will be
	 *  passed in a {@link ThreadEvent} to the {@link ThreadListener#threadError ThreadListener.threadError(ThreadEvent)} method of 
	 *  {@link ThreadListener}s registered for the thread.
	 * <p>
	 *  If this method is called more than once by the same thread, only the last <code>Object</code>
	 *  saved will be returned.
	 *
	 * @param  obj  the <code>Object</code> to be save by the current thread.
	 */
	public static void save(Object obj) {
		if (debug) {
		    System.out.println("ThreadReturn debug: Thread"+Thread.currentThread()+" saving object "+obj);
		}
		threadReturnsMap.put(Thread.currentThread(), obj);
		if (saveImmedeately) {
			if (debug) {
			    System.out.println("ThreadReturn debug: saveImmedeately set calling listeners");
			}
		    // call and remove all the listeners
		    fireListeners(Thread.currentThread());
		}
	}


  /**
   *  Waits for a thread to die and returns the <code>Object</code> it saved.&nbsp;
   *  If the <code>Object</code> saved by the thread, using {@link #save ThreadReturn.save(Object)},
   * is a <code>Throwable</code> object, then this method throws a {@link ThreadException}, otherwise the
   * <code>Object</code> saved is returned by this method.
   * <p>
   * If this method is interrupted, an <code>InterruptedException</code> is thrown.
   *
   * @param  thread                    the thread to wait for. 
	 * @return 	the non-Throwable <code>Object</code> 
	 *      saved by the thread, or <code>null</code> if none saved. 
   *
   * @exception  ThreadInterruptedException  thrown if the exception saved by the thread was
   * <code>InterruptedException</code> or <code>ClosedByInterruptException</code>
   * @exception  ThreadException       thrown if some other <code>Throwable</code> was saved by the thread 
   * @exception  InterruptedException  thrown if the thread calling this method was interrupted.
   * @exception  IllegalThreadStateException thrown if the thread has already died 
   * or has not been started or has been re-started after it has died.
	 */
	public static Object join(Thread thread) throws InterruptedException {
		return join(thread, 0, 0);
	}


  /**
   *  Waits at most <code>milliSecs</code> for a thread to die and returns the <code>Object</code> 
   *  it saved.&nbsp; To wait forever call with milliSecs equal to 0.
   * <p>
   *  If the <code>Object</code> saved by the thread, using {@link #save ThreadReturn.save(Object)},
   * is a <code>Throwable</code> object, then this method throws a {@link ThreadException}, otherwise the
   * <code>Object</code> saved is returned by this method.
   * <p>
   * If this method is interrupted, an <code>InterruptedException</code> is thrown.
   * <p>
   * If this method times out, a {@link TimedOutException} is thrown.
   *
   * @param  thread                    the thread to wait for.
	 * @param  milliSecs                        the milliSecs to wait for the <code>thread</code> to die.
	 * @return 	the non-Throwable <code>Object</code> 
	 *      saved by the thread, or <code>null</code> if none saved. 
   *
   * @exception  ThreadInterruptedException  thrown if the exception saved by the thread was
   * <code>InterruptedException</code> or <code>ClosedByInterruptException</code>
   * @exception  ThreadException       thrown if some other <code>Throwable</code> was saved by the thread 
   * @exception  TimedOutException     thrown if this method timed out.
   * @exception  InterruptedException  thrown if the thread calling this method was interrupted.
   * @exception  IllegalThreadStateException thrown if the thread has already died 
   * or has not been started or has been re-started after it has died.
	 */
	public static Object join(Thread thread, long milliSecs) throws InterruptedException {
		return join(thread, milliSecs, 0);
	}


  /**
   *  Waits at most milliSecs+nanoSecs for a thread to die and returns the <code>Object</code> it saved.&nbsp;
   *  To wait forever call with <code>milliSecs</code> and <code>nanoSecs</code> equal to 0.
   * <p>
   *  If the <code>Object</code> saved by the thread, using {@link #save ThreadReturn.save(Object)},
   * is a <code>Throwable</code> object, then this method throws a {@link ThreadException}, otherwise the
   * <code>Object</code> saved is returned by this method.
   * <p>
   * If this method is interrupted, an <code>InterruptedException</code> is thrown.
   * <p>
   * If this method times out, a {@link TimedOutException} is thrown.
   *
   * @param  thread                    the thread to wait for.
	 * @param  milliSecs                        the milliSecs to wait for the <code>thread</code> to die.
	 * @param  nanoSecs                         the additional nanoSecs to wait (in 
	 *      the range 0-999999).
	 * @return 	the non-Throwable <code>Object</code> 
	 *      saved by the thread, or <code>null</code> if none saved. 
   *
   * @exception  ThreadInterruptedException  thrown if the exception saved by the thread was
   * <code>InterruptedException</code> or <code>ClosedByInterruptException</code>
   * @exception  ThreadException       thrown if some other <code>Throwable</code> was saved by the thread 
   * @exception  TimedOutException     thrown if this method timed out.
   * @exception  InterruptedException  thrown if the thread calling this method was interrupted.
   * @exception  IllegalThreadStateException thrown if the thread has already died 
   * or has not been started or has been re-started after it has died.
	 */
	public static Object join(Thread thread, long milliSecs, int nanoSecs) throws InterruptedException {
		try {
			if (Thread.currentThread().isInterrupted()) {
				// has this main program thread been asked to stop?
				throw new InterruptedException();
			}
			if (hasDied(thread)) {
				// thread has died
				if (thread.isAlive()) {
					// start() was called again
					throw new IllegalThreadStateException(thread.getName() + " was re-started after it had died.");
				}
				// else has died the first time
			} else {
				// thread has not died
				if (!thread.isAlive()) {
					throw new IllegalThreadStateException("Tried to join() " + thread.getName() + " before start() was called.");
				}
			}
			// wait here for the thread to finish
			thread.join(milliSecs, nanoSecs);
			// if an interrupt occurs, let the InterruptedException propagate out of this method
			// check if the thread saved an exception
			Object obj = threadReturnsMap.get(thread);
			if (obj == null) {
				// no exceptions, did it time out?
				if (!hasDied(thread)) {
					throw new TimedOutException(" join(" + thread.getName() + "," + milliSecs + 
							(nanoSecs != 0 ? ("," + nanoSecs) : "") + ") timed out.");
				}
			} else {
				// the thread saved something
				if (obj instanceof Throwable) {
					// re-throw the exception on the main program thread
					if ((obj instanceof InterruptedException) || (obj instanceof ClosedByInterruptException) ){
						throw new ThreadInterruptedException(thread, (Throwable) obj);
					} 
					// else 
					throw new ThreadException(thread, (Throwable) obj);
				}
				// 	else return obj
			}
			return obj;
		} finally {
			thread.interrupt();
			// ask thread to stop if not stopped already
		}
	}


	/**
	 *  This method either just returns if a utility thread is already waiting on 
	 *  this thread or creates an new utility thread This method is called from 
	 *  inside a synchronized(thread) { } block so a listener list for the thread 
	 *  is blocked until this method returns. 
	 *
	 * @param  thread  Description of Parameter 
	 */
	static void joinThread(Thread thread) {
		if (threadJoinMap.get(thread) == null) {
			Thread t = new JoinThread(thread);
			threadJoinMap.put(thread, t);
			t.start();
		}
	}


	/**
	 *  Wait for a thread to finish and then check if any exceptions were 
	 *  thrown.&nbsp; Calls thread listeners when join returns.&nbsp; Waits at most 
	 *  milliSecs+nanoSecs.&nbsp;
	 *  <p>
	 *  To wait forever call with milliSecs and nanoSecs 
	 *  equal to 0.&nbsp; At present only ever called with 0's
	 *  <p>
	 *  If start() was called after the thread died the threadError 
	 *  listener is called with IllegalThreadStateException If start() has not yet 
	 *  been called the threadError listener is called with 
	 *  IllegalThreadStateException.&nbsp; This should not happen. If any 
	 *  threadResult() method throws an exception the corresponding threadError() 
	 *  method is called with that exception. 
	 *
	 * @param  thread     the thread to wait for 
	 * @param  milliSecs  the milliSecs to wait 
	 * @param  nanoSecs   the additional nanoSecs to wait (in the range 0-999999) 
	 */
	static void listenerJoin(Thread thread, long milliSecs, int nanoSecs) {
		try {
			if (hasDied(thread)) {
				// thread has died
				if (thread.isAlive()) {
					// start() was called again
					// call error listener with
					fireThreadError(thread, new ThreadEvent(thread, 
							new IllegalThreadStateException(thread.getName() + " was re-started after it had died.")));
				}
				// else has died the first time
			} else {
				// thread has not died
				if (!thread.isAlive()) {
					// call error listener with
					fireThreadError(thread, new ThreadEvent(thread, 
							new IllegalThreadStateException("Tried to join() " + thread.getName() + " before start() was called.")));
					// call all listeners
				}
			}
			// wait here for the thread to finish
			try {
				thread.join(milliSecs, nanoSecs);
			} catch (InterruptedException ex) {
				// call interrupted listener
				fireThreadError(thread, new ThreadEvent(thread, ex));
				// call all listeners
			}
			if (!hasDied(thread)) {
				// call timedOut listener
				fireThreadError(thread,
						new ThreadEvent(thread, new TimedOutException(thread.getName())));
				// call all listeners
			} else {
				// check if the thread saved an object
			    // fire listeners for this return
			    fireListeners(thread);
			}
		} catch (Throwable t) {
			// should not normally get here.
			System.out.println(" caught throwable "+StackTrace.toString(t));
			System.out.println(" thread "+thread.getName());
			fireThreadError(thread, new ThreadEvent(thread, t));
		}
	}


	/**
	 *  Fires the appropiate listeners with the current thread saved object and then removes them
	 *
	 * @param  thread  the thread whos listeners are to be called
	 */
	private static void fireListeners(Thread thread) {
		// check if the thread saved an object
		Object obj = threadReturnsMap.get(thread);
		if (obj == null) {
			// no errors or returns
			fireThreadReturn(thread, new ThreadEvent(thread, obj));
		} else if (!(obj instanceof Throwable)) {
			// a return
			fireThreadReturn(thread, new ThreadEvent(thread, obj));
		} else {
			// call error or stopped listener
			if ((obj instanceof InterruptedException) || (obj instanceof ClosedByInterruptException) ){
				fireThreadInterrupted(thread, new ThreadEvent(thread, (Throwable) obj));
			} else {
				// an error
				fireThreadError(thread, new ThreadEvent(thread, (Throwable) obj));
			}	
		}
	}
	
	/**
	 *  Call all threadResult methods for this thread
	 *  Clears the
	 *  list of listeners for this thread.
	 *
	 * @param  thread  the thread whos listeners are being called
	 * @param  event  the ThreadEvent to pass to threadResult
	 */
	private static void fireThreadReturn(Thread thread, ThreadEvent event) {
		ThreadListener[] listeners = new ThreadListener[0];
		synchronized (thread) {
			// lock this thread's listeners against modification by other threads
			ArrayList listenerList = (ArrayList) threadListenerMap.get(thread);
			if (listenerList == null) {
				return; // no listeners
			}
			ArrayList calledListenerList = (ArrayList) calledListenerMap.get(thread);
			if (calledListenerList == null) {
				calledListenerList = new ArrayList();
			}
		  listeners = (ThreadListener[])(listenerList.toArray(new ThreadListener[listenerList.size()]));
		  // add these listeners to the called list
		  for (int i=0; i<listeners.length; i++) {
		  	calledListenerList.add(listeners[i]);
		  }
			// clear out listeners
			threadListenerMap.put(thread, null);
			// update called listeners
			calledListenerMap.put(thread,calledListenerList);
			
			// clear out the ThreadJoin map also
			threadJoinMap.put(thread, null);
		}	
		// call all the threadResult methods
		// catch errors and call threadError
		for (int i= 0; i< listeners.length; i++) {
			try{
				listeners[i].threadResult(event);
			} catch (Throwable t) {
				try {
					listeners[i].threadError(new ThreadEvent(event.getThread(), t));
				} catch (Throwable et) {
					// ignore
				}	
			}	
		}
	}


	/**
	 *  Call all threadInterrupted methods for this thread
	 *  Clears the
	 *  list of listeners for this thread.
	 *
	 * @param  thread  the thread whos listeners are being called
	 * @param  event  the ThreadEvent to pass to threadResult
	 */
	private static void fireThreadInterrupted(Thread thread, ThreadEvent event) {
		ThreadListener[] listeners = new ThreadListener[0];
		synchronized (thread) {
			// lock this thread's listeners against modification by other threads
			ArrayList listenerList = (ArrayList) threadListenerMap.get(thread);
			if (listenerList == null) {
				return; // no listeners
			}
			ArrayList calledListenerList = (ArrayList) calledListenerMap.get(thread);
			if (calledListenerList == null) {
				calledListenerList = new ArrayList();
			}
		  listeners = (ThreadListener[])(listenerList.toArray(new ThreadListener[listenerList.size()]));
		  // add these listeners to the called list
		  for (int i=0; i<listeners.length; i++) {
		  	calledListenerList.add(listeners[i]);
		  }
			// clear out listeners
			threadListenerMap.put(thread, null);
			// update called listeners
			calledListenerMap.put(thread,calledListenerList);
			
			// clear out the ThreadJoin map also
			threadJoinMap.put(thread, null);
		}	
		// call all the threadStopped methods
		// catch errors and call threadError
		for (int i= 0; i< listeners.length; i++) {
			try{
				listeners[i].threadInterrupted(event);
			} catch (Throwable t) {
				try {
					listeners[i].threadError(new ThreadEvent(event.getThread(), t));
				} catch (Throwable et) {
					// ignore
				}	
			}	
		}
	}
	
	
	/**
	 *  Call all threadError methods for this thread.
	 *  Clears the
	 *  list of listeners for this thread.
	 *
	 * @param  thread  the thread whos listeners are being called
	 * @param  event  the ThreadEvent to pass to threadError
	 */
	private static void fireThreadError(Thread thread, ThreadEvent event) {
		ThreadListener[] listeners = new ThreadListener[0];
		synchronized (thread) {
			// lock this thread's listeners against modification by other threads
			ArrayList listenerList = (ArrayList) threadListenerMap.get(thread);
			if (listenerList == null) {
				return; // no listeners
			}
			ArrayList calledListenerList = (ArrayList) calledListenerMap.get(thread);
			if (calledListenerList == null) {
				calledListenerList = new ArrayList();
			}
		  listeners = (ThreadListener[])(listenerList.toArray(new ThreadListener[listenerList.size()]));
		  // add these listeners to the called list
		  for (int i=0; i<listeners.length; i++) {
		  	calledListenerList.add(listeners[i]);
		  }
			// clear out listeners
			threadListenerMap.put(thread, null);
			// update called listeners
			calledListenerMap.put(thread,calledListenerList);
			
			// clear out the ThreadJoin map also
			threadJoinMap.put(thread, null);
		}	
		// call all the threadErrors methods
		// catch errors and ignore
		for (int i= 0; i< listeners.length; i++) {
			try {
				listeners[i].threadError(event);
			} catch (Throwable t) {
				// ignore
			}	
		}
	}

}

/**
 *  Monitor Thread Class 
 *
 * @author     matthew ford 
 * @created    5 May 2002 
 */
class JoinThread extends Thread {


	/**
	 *  The thread to monitor 
	 */

	Thread threadToJoin;


	/**
	 *  Constructor for the JoinThread object 
	 *
	 * @param  thread  the thread to monitor 
	 */
	JoinThread(Thread thread) {
		threadToJoin = thread;
	}


	/**
	 *  Main processing method for the JoinThread object 
	 */
	public void run() {
		// wait forever for thread to die.
		ThreadReturn.listenerJoin(threadToJoin, 0, 0);
	}
}